@@ -10,7 +10,7 @@
 #  ┃ of the [Apache License 2.0](https://www.apache.org/licenses/LICENSE-2.0). ┃
 #  ┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┛
 
-__version__ = "3.6.1"
+__version__ = "3.7.0"
 __all__ = [
     "_jupyter_labextension_paths",
     "Server",
@@ -18,6 +18,9 @@ __all__ = [
     "PerspectiveError",
     "ProxySession",
     "AsyncClient",
+    "AsyncServer",
+    "num_cpus",
+    "set_num_cpus",
 ]
 
 import functools
@@ -26,12 +29,15 @@ from .perspective import (
     Client,
     PerspectiveError,
     ProxySession,
-    PySyncServer as Server,
+    PyServer as Server,
+    PyAsyncServer as AsyncServer,
     AsyncClient,
     # NOTE: these are classes without constructors,
     # so we import them just for type hinting
     Table,  # noqa: F401
     View,  # noqa: F401
+    num_cpus,
+    set_num_cpus,
 )
 
 
